/*

 Provide "int limits[MAXMARKS]" to the rest of the algorithm
 Uses: MAXMARKS, maxmarks, maxlength, depth, dist[][0]

 limits.h broken out of the main halfTweakedGARSP.c files
     Added OGRdist (32bit only for now, but could/should later be extended)
     Wrote optimized versions for KTEST=0, KTEST=1 and KTEST>1
     Use choose in the midpoint to improve the halflength limit (andreasb)

 limits_steps.h
     Include the calculated limit in the dists and see if the limit can be reduced even further (tempo)
     Also include the new DIST that would be added if we would set a mark at tempo (e.g. DIST@(depth+1))
     Add a while loop to reduce limit if there wouldnt be sapce for a mark at that place in COMP
 
 */

/* For OGR-X you should select chooseV3 with CHOOSEMARKS <= X-3 (because steps is better than CHOOSE in the early stages) */
/* It is desireable to use a chooseV3 where (CHOOSEMARKS-1)*2 > X-3 (to ensure performance of constructed double CHOOSE)*/
/* You should have at least a CHOOSEMARKS that is > x/4 (otherwise the usefulness of CHOOSE to the midpoint decreases) */
/* Also note that CHOOSEMARKS has to fullfill the requirements of KTEST! */

/* e.g. to search OGR-14 you should have CHOOSEMARKS <= 11 and at least CHOOSEMARKS >= 4 */
/* good values would also fulfill (CHOOSEMARKS-1)*2 > 11 --> 7,8,9,10,11 (more is better)*/

#define MINIMUM(x,y) ((x<y) ? (x) : (y))
#define KTEST 1  /* NOTE: KTEST restricts your ability to find very short OGRs
                         e.g. using KTEST=0 you can find all OGRs (except OGR-0 though OGR-2 but these are trivial)
                              using KTEST=1 you cannot reliably find OGRs <= OGR-4
                              using KTEST=2 you cannot reliably find OGRs <= OGR-6 */
/*
   Use KTEST=0 for <= OGR-14
   Use KTEST=1 for >= OGR-15 (reduces the nodecount!!!)
   Use KTEST=2 ???
*/

#if (KTEST>0)

#if (CHOOSEMARKS<=(KTEST*2)+1)
/* calculation of halflength requires (CHOOSEMARKS>halfdepth2-halfdepth1) */
/* e.g.: COOSEMARKS > 3 for KTEST 1, > 5 for KTEST 2, > 7 for KTEST 3,... */
#error please use a choosefile with more CHOOSEMARKS or reduce KTEST
#endif


#define SETUP_LIMITS                                                                 			\
const int OGR[32] = { /*  1 */    0,   1,   3,   6,  11,  17,  25,  34,  44,  55,  72,			\
                      /* 12 */   85, 106, 127, 151, 177, 199, 216, 246, 283, 333, 356,			\
                      /* 23 */  372, 425, 480, 492, 553, 585, 623, 680, 747, 784 };			\
const int halfdepth1 = (maxmarks/2)-KTEST;								\
const int halfdepth2 = (maxmarks/2)+(maxmarks%2)+KTEST;   						\
signed int limit[MAXMARKS]; /*current limit, populated by limits.h (based on CHOOSE, MIDPOINT, etc.) */


#define CALCULATE_LIMITS                                                                                \
                                                                                                        \
/* CHOOSE */                                                                                            \
limit[depth] = maxlength-chooseV3[GET_DIST_CHOOSEBITS][(maxmarks-depth)];                               \
/*if(limit[depth]<=count[depth]) {break;}*/                                                             \
                                                                                                        \
/* MIDPOINT */                                                                                          \
if (depth <= halfdepth2) {                                                                              \
    if (depth <= halfdepth1) {                                                                          \
        int tempLimit;                                                                                  \
        int halflength = (((maxlength-1)-chooseV3[GET_DIST_CHOOSEBITS][(halfdepth2-halfdepth1)])/2);    \
                                                                                                        \
        if ((count[depth]+chooseV3[GET_DIST_CHOOSEBITS][(maxmarks-(depth-1))])>maxlength) { break; }    \
        if ((count[depth]+chooseV3[GET_DIST_CHOOSEBITS][(halfdepth1-(depth-1))])>halflength) { break; } \
                                                                                                        \
        tempLimit = halflength-chooseV3[GET_DIST_CHOOSEBITS][(halfdepth1-depth)];                       \
                                                                                                        \
        limit[depth]=MINIMUM(limit[depth],tempLimit);                                                   \
                                                                                                        \
        /* Announce where we are */                                                                     \
                if ((depth<=5)&&(depth>1)) {                                                            \
                    int i;                                                                              \
                    printf("STATUS: currently working at");                                             \
                    for(i=1; i<depth; i++ ) printf("%4d",count[i]-count[i-1]);                          \
                    printf(" limited at %d\n",limit[depth-1]-count[i-2]); fflush(0);                    \
                }                                                                                       \
                                                                                                        \
        if (maxlength-count[depth]<=finalizeV4_max[maxmarks-(depth-1)]) {                                                           \
            signed int row=finalizeV4_rows[maxmarks-(depth-1)][(maxlength-count[depth])-(OGR[maxmarks-(depth-1)])]-1;               \
            /*                                                                                                                      \
            printf("maxlen-count[depth]=%d finalizeV4_rows=%d\n",maxlength-count[depth],row);                                       \
            printf("count[depth]=%d finalizeV4_max[maxmarks-(depth-1)]=%d\n",count[depth],finalizeV4_max[maxmarks-(depth-1)]);      \
            fflush(0);                                                                                                              \
            */                                                                                                                      \
            printf("maxlen-%d=%d units for GR-%d (%d rows)\n",count[depth],maxlength-count[depth],(maxmarks-(depth-1))+1,row+1);    \
            fflush(0);                                                                                                              \
                                                                                                                                    \
            for (;row>=0;row--) {                                                                                                   \
                if (((GET_DIST_FINALIZEBITS)&(finalizeV4_DIST[(maxmarks-(depth-1))-FINOFFSET][row]))==0) {                          \
                    /* this could be enhanced by using row as an identifier instead of 0 and 1 */                                   \
                    uint16_t diffs[((((NUMWORDS*32*2)+1)+(NUMDIFFS*2))+1)]={0}; /* the number of diffs supported, plus 1 */	        \
                    int i,j,slider,valid = 1;    									                                                \
                    /*printf("maxmarks-(depth-1)=%d row=%d\n",maxmarks-(depth-1),row);fflush(0);*/                                  \
                    /* create full ruler */									                                                        \
                    for (i=0;i<=maxmarks-depth;i++) {							                                                    \
                        count[depth+i]=count[(depth-1)+i]+finalizeV4[(maxmarks-(depth-1))-FINOFFSET][row][i];		                \
                    }											                                                                    \
                    /* verify golombness, lower left to upper right */					                                            \
                    for (i=1; i <= maxmarks; i++) {								                                                    \
                        for (j=0; j<i; j++) {								                                                        \
                            int diff = count[i]-count[j];							                                                \
                            if (diffs[diff]==1) { i=maxmarks; valid=0; break; }				                                        \
                            diffs[diff] = 1;								                                                        \
                        }											                                                                \
                    }											                                                                    \
                                                                                                                                    \
/*                                                                                                                                  \
 for (slider=32-MY_CHOOSE_BITS;slider<=31;slider++) {                                                                               \
    if ((0x01<<slider)&(global_bitmap) != 0) {                                                                                      \
        if (diffs[32-slider]==1) { slider=32; valid=0; break; }                                                                     \
    }                                                                                                                               \
}				 	                                                                                                                \
if ((valid)&&(count[maxmarks]<=maxlength)) {			                                                                            \
       printf("maxlength from %d reduced to %d!\n",maxlength,count[maxmarks]-1);fflush(0);                                          \
       maxlength=count[maxmarks]-1;			                                                                                        \
}							                                                                                                        \
*/                                                                                                                                  \
                    if (valid) { printf("FOUND A RULER by construction (row=%d)!!!\n",row);			                                \
                        printf("{%3d",count[maxmarks]);                                                                             \
                        if ((count[maxmarks]-count[maxmarks-1])>count[1]) {                                                         \
                            for(i=1; i<=maxmarks; i++ ) printf(",%d",count[i]-count[i-1]); printf("},\n");                          \
                        } else {                                                                                                    \
                            for(i=maxmarks; i>0; i-- ) printf(",%d",count[i]-count[i-1]); printf("},\n");                           \
                        }                                                                                                           \
                    fflush(0); }		                                                                                            \
                                                                                                                                    \
                                                                                                                                    \
                    valid = 1;          									                                                        \
                    /* create full mirror ruler */									                                                \
                    for (i=0;i<=maxmarks-depth;i++) {							                                                    \
                        count[depth+i]=count[(depth-1)+i]+finalizeV4[(maxmarks-(depth-1))-FINOFFSET][row][(maxmarks-depth)-i];		\
                    }											                                                                    \
                    /* verify golombness, lower left to upper right */					                                            \
                    for (i=1; i <= maxmarks; i++) {								                                                    \
                        for (j=0; j<i; j++) {								                                                        \
                            int diff = count[i]-count[j];							                                                \
                            if (diffs[diff]==2) { i=maxmarks; valid=0; break; }				                                        \
                            diffs[diff] = 2;								                                                        \
                        }											                                                                \
                    }											                                                                    \
                                                                                                                                    \
/*                                                                                                                                  \
 for (slider=32-MY_CHOOSE_BITS;slider<=31;slider++) {                                                                               \
    if ((0x01<<slider)&(global_bitmap) != 0) {                                                                                      \
        if (diffs[32-slider]==2) { slider=32; valid=0; break; }                                                                     \
    }                                                                                                                               \
}				 	                                                                                                                \
if ((valid)&&(count[maxmarks]<=maxlength)) {			                                                                            \
       printf("maxlength from %d reduced to %d!\n",maxlength,count[maxmarks]-1);fflush(0);                                          \
       maxlength=count[maxmarks]-1;			                                                                                        \
}							                                                                                                        \
*/                                                                                                                                  \
                    if (valid) { printf("FOUND A RULER by construction (row=%d)!!!\n",row);			                                \
                        printf("{%3d",count[maxmarks]);                                                                             \
                        if ((count[maxmarks]-count[maxmarks-1])>count[1]) {                                                         \
                            for(i=1; i<=maxmarks; i++ ) printf(",%d",count[i]-count[i-1]); printf("},\n");                          \
                        } else {                                                                                                    \
                            for(i=maxmarks; i>0; i-- ) printf(",%d",count[i]-count[i-1]); printf("},\n");                           \
                        }                                                                                                           \
                    fflush(0); }		                                                                                            \
                }                                                                                                                   \
                                                                                                                                    \
                                                                                                                                    \
            }												                                                                        \
            /* restoring already done ?! 									                                                        \
            count[depth]=count[depth-1];  is the recreation of count[depth] really necessary? */                                    \
            break;	                                                                                                                \
        }													                                                                        \
                                                                                                                                    \
                                                                                                        \
                                                                                                        \
    } else {												                                            \
        const int tempLimit = ((maxlength-1)-count[halfdepth1])-(chooseV3[GET_DIST_CHOOSEBITS][(halfdepth2-depth)]);	\
        limit[depth]=MINIMUM(limit[depth],tempLimit);							                        \
    }													                                                \
}


#define RESET_LIMITS											\
{ int store_depth = depth; 										\
  for (depth=1; depth<store_depth;depth++) { CALCULATE_LIMITS }						\
  depth = store_depth; }

#else // (KTEST==0)


#define SETUP_LIMITS                                                                 			\
const int OGR[32] = { /*  1 */    0,   1,   3,   6,  11,  17,  25,  34,  44,  55,  72,			\
                      /* 12 */   85, 106, 127, 151, 177, 199, 216, 246, 283, 333, 356,			\
                      /* 23 */  372, 425, 480, 492, 553, 585, 623, 680, 747, 784 };			\
const int halfdepth1 = (maxmarks/2);									\
int halflength = (!(maxmarks%2) && (maxlength%2)) ? (maxlength/2) : (maxlength/2)-1;			\
int limit[MAXMARKS];	/* current limit, populated by limits.h (based on CHOOSE, MIDPOINT, etc.) */


#define CALCULATE_LIMITS	                 							\
/* CHOOSE */												\
if (CHOOSEMARKS>maxmarks-depth) {									\
    limit[depth] = maxlength-chooseV3[GET_DIST_CHOOSEBITS][(maxmarks-depth)];				\
} else {												\
    limit[depth] = maxlength-OGR[maxmarks-depth]; /* midpoint reduction may supercede this */		\
}													\
/* MIDPOINT */												\
if (depth<=halfdepth1) {										\
    int tempLimit;											\
    if (CHOOSEMARKS>halfdepth1-depth) {									\
        tempLimit = halflength-chooseV3[GET_DIST_CHOOSEBITS][(halfdepth1-depth)];			\
    } else {												\
        tempLimit = halflength-OGR[halfdepth1-depth];							\
    }													\
    limit[depth]=MINIMUM(limit[depth],tempLimit);							\
                                                                                                        \
} else if ((depth==(halfdepth1+1)) && (maxmarks%2)) {							\
    const int tempLimit = maxlength-1-count[halfdepth1];						\
    limit[depth]=MINIMUM(limit[depth],tempLimit);							\
    /*printf("l%d r%d\n",count[halfdepth1],maxlength-1-count[halfdepth1]);*/				\
}


#define RESET_LIMITS											\
{ int store_depth = depth; 										\
  int halflength = (!(maxmarks%2) && (maxlength%2)) ? (maxlength/2) : (maxlength/2)-1;			\
  for (depth=1; depth<store_depth;depth++) { CALCULATE_LIMITS }						\
  depth = store_depth; }


#endif
